((nil . ((indent-tabs-mode . nil)
	 (tab-width . 8)
	 (fill-column . 79)
	 (ispell-local-dictionary . "american")))
 (c-mode . ((c-file-style . "bsd")
	    (c-basic-offset . 2)))
 (java-mode . ((c-basic-offset . 4)))
 (julia-mode . ())
 (makefile-gmake-mode . ((indent-tabs-mode . t)))
 (sh-mode . ((sh-basic-offset . 4)))
 (tcl-mode . ((tcl-default-application . "wish")
	      (tcl-indent-level . 2)))
 (yorick-mode . ((c-basic-offset . 2)))
)
